<?php

namespace Drupal\media_migration;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\field\Plugin\migrate\source\d7\Field;
use Drupal\field\Plugin\migrate\source\d7\FieldInstance;
use Drupal\field\Plugin\migrate\source\d7\ViewMode;
use Drupal\migrate\Exception\RequirementsException;
use Drupal\migrate\Plugin\MigrateSourcePluginManager;
use Drupal\migrate\Plugin\MigrationDeriverTrait;
use Drupal\migrate\Plugin\MigrationPluginManager;
use Drupal\migrate\Row;
use Drupal\migrate_drupal\Plugin\migrate\FieldMigration;
use Drupal\migrate_drupal\Plugin\migrate\source\DrupalSqlBase;
use Psr\Log\LoggerInterface;

/**
 * MigratePluginAlterer service.
 */
class MigratePluginAlterer {

  use MigrationDeriverTrait;

  /**
   * The plugin.manager.migration service.
   *
   * @var \Drupal\migrate\Plugin\MigrationPluginManager
   */
  protected $pluginManagerMigration;

  /**
   * The Migrate source plugin manager service.
   *
   * @var \Drupal\migrate\Plugin\MigrateSourcePluginManager
   */
  protected $sourceManager;

  /**
   * The plugin.manager.media_wysiwyg service.
   *
   * @var \Drupal\media_migration\MediaWysiwygPluginManager
   */
  protected $pluginManagerMediaWysiwyg;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The type of the embed token transformation.
   *
   * This is 'entity_embed' or NULL, depending on whether the destination
   * module is available or not.
   *
   * @var string|null
   */
  protected $embedTokenTransformType;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a MigratePluginAlterer object.
   *
   * @param \Drupal\migrate\Plugin\MigrationPluginManager $plugin_manager_migration
   *   The migration plugin manager.
   * @param \Drupal\migrate\Plugin\MigrateSourcePluginManager $source_manager
   *   The Migrate source plugin manager.
   * @param \Drupal\media_migration\MediaWysiwygPluginManager $plugin_manager_media_wysiwyg
   *   The Media WYSIWYG plugin manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(MigrationPluginManager $plugin_manager_migration, MigrateSourcePluginManager $source_manager, MediaWysiwygPluginManager $plugin_manager_media_wysiwyg, LoggerInterface $logger, ModuleHandlerInterface $module_handler, EntityTypeManagerInterface $entity_type_manager) {
    $this->pluginManagerMigration = $plugin_manager_migration;
    $this->sourceManager = $source_manager;
    $this->pluginManagerMediaWysiwyg = $plugin_manager_media_wysiwyg;
    $this->logger = $logger;
    $this->moduleHandler = $module_handler;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Alters migrate plugins.
   *
   * @param array $migrations
   *   The array of migration plugins.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a plugin cannot be found.
   */
  public function alter(array &$migrations) {
    $this->alterFieldMigrations($migrations);
    $this->addMediaWysiwygProcessor($migrations);
    $this->alterFilterFormatMigration($migrations);
  }

  /**
   * Alters field migrations from file_entity/media in 7 to media in 8.
   *
   * @param array $migrations
   *   The array of migration plugins.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If a plugin cannot be found.
   */
  protected function alterFieldMigrations(array &$migrations) {
    // Collect all file entity -> media entity migration which have additional
    // fields. Although these migrations depend on the "d7_field_instance"
    // migration (and implicitly on "d7_field" migration), it isn't enough:
    // the parent media type entity has to exist prior its all field storage
    // entity being created, so "d7_field" (and maybe "d7_field_instance") has
    // to depend on the media type migrations which have additional fields.
    $media_types_with_fields_migration_ids = array_reduce($migrations, function (array $carry, array $migration_def) {
      $deps_required = isset($migration_def['migration_dependencies']['required'])
        ? $migration_def['migration_dependencies']['required']
        : [];

      if (!in_array('d7_field_instance', $deps_required, TRUE)) {
        return $carry;
      }

      foreach ($deps_required as $requirement) {
        $requirement_parts = explode(':', $requirement);
        if ($requirement_parts[0] === 'd7_file_entity_type') {
          $carry[$requirement] = $requirement;
        }
      }
      return $carry;
    }, []);

    foreach ($migrations as &$migration) {
      $migration_stub = $this->pluginManagerMigration->createStubMigration($migration);
      $source = NULL;
      $configuration = $migration['source'];
      if (!empty($migration['source']['plugin'])) {
        $source = $this->sourceManager->createInstance($migration['source']['plugin'], $configuration, $migration_stub);
        if (is_a($migration['class'], FieldMigration::class, TRUE)) {

          // Field storage, instance, widget and formatter migrations.
          if (is_a($source, Field::class) || is_a($source, FieldInstance::class)) {
            $this->mapFileFields($migration);
          }
        }

        // View Modes.
        if (is_a($source, ViewMode::class)) {
          $this->mapFileToMediaBundle($migration);
        }

        // D7 field storage and field instance migrations should depend on
        // media types which have extra fields, because field storage config
        // entities require their host entity.
        $field_storage_and_instance_source_plugin_ids = [
          'd7_field',
          'd7_field_instance',
        ];
        if (
          in_array($migration['source']['plugin'], $field_storage_and_instance_source_plugin_ids) &&
          !empty($media_types_with_fields_migration_ids)
        ) {
          $required_migration_deps = isset($migration['migration_dependencies']['required'])
            ? $migration['migration_dependencies']['required']
            : [];
          $migration['migration_dependencies']['required'] = array_unique(
            array_merge(
              $required_migration_deps,
              array_values($media_types_with_fields_migration_ids)
            )
          );
        }
      }
    }
  }

  /**
   * Appends a processor to transform media_wysiwyg tokens to entity_embeds.
   *
   * Find field instances with text processing and pass them to a
   * MediaWysiwyg plugin that will add processors to the respective
   * migrations.
   *
   * @see \Drupal\media_migration\Plugin\MediaWysiwyg\Node
   */
  protected function addMediaWysiwygProcessor(array &$migrations) {
    $field_instance_migration = static::getSourcePlugin('d7_field_instance');

    try {
      assert($field_instance_migration instanceof DrupalSqlBase);
      $field_instance_migration->checkRequirements();
    }
    catch (RequirementsException $e) {
      // This exception happens when we run migrations with migrate:import or
      // check the status with migrate:update. We just we need to run the
      // code below when we are generating migrations with migrate:upgrade so
      // it is safe to just return here if we reach to this point.
      return;
    }

    $entity_destination_plugins = [
      'entity',
      'entity_revision',
      'entity_complete',
      'entity_reference_revisions',
    ];
    $content_entity_migrations = array_filter($migrations, function (array $migration) use ($entity_destination_plugins) {
      if (!in_array('Drupal 7', $migration['migration_tags'] ?? [], TRUE)) {
        return FALSE;
      }
      $destination_parts = explode(':', $migration['destination']['plugin']);
      if (
        count($destination_parts) !== 2 ||
        !in_array($destination_parts[0], $entity_destination_plugins)
      ) {
        return FALSE;
      }
      $destination_entity_type = $destination_parts[1];
      $definition = $this->entityTypeManager->getDefinition($destination_entity_type, FALSE);
      if ($definition instanceof ContentEntityTypeInterface) {
        return TRUE;
      }

      return FALSE;
    });
    $content_entity_destinations = [];
    foreach ($content_entity_migrations as $migration_plugin_id => $content_entity_migration) {
      $destination_entity_type = explode(':', $content_entity_migration['destination']['plugin'])[1];
      $content_entity_destinations[$destination_entity_type][] = $migration_plugin_id;
    }

    foreach ($field_instance_migration as $row) {
      assert($row instanceof Row);
      if ($row->getSourceProperty('settings/text_processing') === 1) {
        $entity_type_id = $row->getSourceProperty('entity_type');
        $plugin_id = array_key_exists($entity_type_id, $content_entity_destinations) ? 'fallback' : $entity_type_id;

        try {
          $plugin = $this->pluginManagerMediaWysiwyg->createInstance($plugin_id);
          assert($plugin instanceof MediaWysiwygInterface);
          $migrations = $plugin->process($migrations, $row, $content_entity_destinations[$entity_type_id] ?? []);
          continue;
        }
        catch (PluginException $e) {
        }
        $this->logger->warning(sprintf('Could not find a MediaWysiwyg plugin "%s" for field "%s". You probably need to create a new one. Have a look at \Drupal\media_migration\Plugin\MediaWysiwyg\FieldCollectionItem for an example.', $entity_type_id, $row->getSourceProperty('field_name')));
      }
    }
  }

  /**
   * Maps Drupal 7 media_filter filter plugin to a Drupal 8|9 filter plugin.
   *
   * If Entity Embed module is installed on the destination site, this method
   * maps the media_embed filter plugin (Drupal 7 Media WYSIWYG module) to
   * entity_embed filter plugin (from the Entity Embed module).
   * If Entity Embed is unavailable, the media_filter filter will be mapped to
   * media_embed filter (from core Media Library module).
   *
   * @param array $migrations
   *   The array of migration plugins.
   */
  protected function alterFilterFormatMigration(array &$migrations) {
    $destination_filter_plugin_id = MediaMigration::getEmbedTokenDestinationFilterPlugin();
    // If entity_embed is not installed, the destination entity type of the
    // "d7_embed_button_media" migration is missing.
    if (!$this->moduleHandler->moduleExists('entity_embed')) {
      unset($migrations['d7_embed_button_media']);
    }

    if (isset($migrations['d7_filter_format']) || MediaMigration::embedTokenDestinationFilterPluginIsValid($destination_filter_plugin_id)) {
      $migrations['d7_filter_format']['process']['filters']['process']['id']['map']['media_filter'] = $destination_filter_plugin_id;
    }
    else {
      // We don't know the transform type or the filter format migration does
      // not exist.
      return;
    }

    if (MediaMigration::MEDIA_TOKEN_DESTINATION_FILTER_ENTITY_EMBED == $destination_filter_plugin_id && isset($migrations['d7_filter_format']) && isset($migrations['d7_embed_button_media'])) {
      $migrations['d7_filter_format']['migration_dependencies']['required'][] = 'd7_embed_button_media';
    }

    // We have to add <drupal-entity> or <drupal-media> to the allowed html
    // tag's list.
    if (isset($migrations['d7_filter_format'])) {
      $filter_plugin_settings_processes = $this->makeAssociative($migrations['d7_filter_format']['process']['filters']['process']['settings']);
      $filter_plugin_settings_processes[] = [
        'plugin' => 'filter_settings_embed_media',
      ];
      $migrations['d7_filter_format']['process']['filters']['process']['settings'] = $filter_plugin_settings_processes;
    }
  }

  /**
   * Maps file type fields to media ones.
   *
   * @param array $migration
   *   The migration to alter.
   */
  protected function mapFileFields(array &$migration) {
    $entity_type_process = $migration['process']['entity_type'];

    if (isset($entity_type_process['media_migration'])) {
      return;
    }
    $entity_type_process = $this->makeAssociative($entity_type_process);

    $entity_type_process['media_migration'] = [
      'plugin' => 'static_map',
      'map' => [
        'file' => 'media',
      ],
      'bypass' => TRUE,
    ];
    $migration['process']['entity_type'] = $entity_type_process;
  }

  /**
   * Adds static mapping from file to media to a entity references.
   *
   * @param array $migration
   *   The migration to alter.
   */
  protected function mapFileToMediaBundle(array &$migration) {
    $entity_type_process = $migration['process']['targetEntityType'];

    if (isset($entity_type_process['media_migration'])) {
      return;
    }
    $entity_type_process = $this->makeAssociative($entity_type_process);

    $entity_type_process['media_migration'] = [
      'plugin' => 'static_map',
      'map' => [
        'file' => 'media',
      ],
      'bypass' => TRUE,
    ];
    $migration['process']['targetEntityType'] = $entity_type_process;
  }

  /**
   * Ensures that a plugin process mapping is an associative array.
   *
   * @param array|string $plugin_process
   *   The plugin process mapping.
   *
   * @return array
   *   The plugin process mapping as an associative array.
   */
  protected function makeAssociative($plugin_process) {
    if (!is_array($plugin_process)) {
      $plugin_process = [
        [
          'plugin' => 'get',
          'source' => $plugin_process,
        ],
      ];
    }
    elseif (array_key_exists('plugin', $plugin_process)) {
      $plugin_process = [$plugin_process];
    }

    return $plugin_process;
  }

}
